/* shape.c
 *
 * Functions for manipulating shapes
 */
#include "sim.h"
#include "server.h"

static hashtable st = {0};

static int snum = 0;	/* for unique shape names */


/* shape_name
 * Returns a unique shape name string
 */
char *shape_name(void)
{
static char nbuf[256];

sprintf(nbuf,"sh%d.%s",snum++,findmyhostname());
return nbuf;
}


/* create_shape
 * Allocate and init a new shape. Give it a unique name in case none is 
 * specified when parsing
 */
shape *create_shape(void)
{
shape *ns = scalloc(1,sizeof(shape));

ns->pos.x = ns->pos.y = ns->pos.z = 0.0;
ns->hidden = 0;
ns->rota = 0;
ns->rotx = ns->roty = ns->rotz = 0.0;
ns->scx = ns->scy = ns->scz = 1.0;
return ns;
}


/* Hash table functions
 */
void add_shape(shape *s)
{ add_entry(&st,s); }

shape *lookup_shape(char *name)
{ return lookup_entry(&st,name); }

void reset_shape(void)
{ reset_position(&st); }

shape *next_shape(void)
{ return next_position(&st); }

void remove_shape(shape *sh)
{ remove_entry(&st,sh); }


/* free_shape
 * Deep-free a shape structure.
 */
void free_shape(shape *sh)
{
int i;
shape *sp, *nxt;

switch(sh->type) {
case st_poly:
	for(i=0; i<sh->shp.poly.facec; i++)
		free(sh->shp.poly.pt[i]);
	break;
case st_coll:
	for(sp=sh->shp.coll.head; sp; sp=nxt) {
		nxt = sp->next;
		free_shape(sp);
		}
	break;
	} /* end switch */
free(sh);
}


/* move_shape
 * Move a 3d shape by rx,ry,rz
 */
void move_shape(shape *s, double rx, double ry, double rz)
{
s->pos.x += rx;
s->pos.y += ry;
s->pos.z += rz;
}


/* shade_shape
 * Change the surface type of a shape and all it's children
 */
void shade_shape(shape *s, int shd)
{
int i;

s->surf = shd;
if (s->type == st_coll) {
	shape *p;
	for(p=s->shp.coll.head; p; p=p->next)
		shade_shape(p,shd);
	}
}


/* copy_shape
 * Create a shape which is a copy of another
 */
shape *copy_shape(shape *src)
{
shape *ns = create_shape();

ns->colour = src->colour;
ns->pos    = src->pos;
ns->hidden = src->hidden;
ns->rota   = src->rota; ns->rotx   = src->rotx;
ns->roty   = src->roty; ns->rotz   = src->rotz; 
ns->scx    = src->scx;  ns->scy    = src->scy; 
ns->scz    = src->scz; 
ns->surf   = src->surf;
ns->tex    = src->tex;
if (ns->tex.data)
    ns->tex.data = memdupe(ns->tex.data,ns->tex.w * ns->tex.h * 3);
if (src->cmt)
    ns->cmt = strdupe(src->cmt);
ns->type = st_copy;
ns->shp.copy.src = src;
return ns;
}


/* deep_copy_shape
 * Fully duplicate a 3d shape
 */
shape *deep_copy_shape(shape *s)
{
shape *d = memdupe(s,sizeof(shape));
shape *sp;
int i;

d->name[0] = 0;
d->next = NULL;
if (s->tex.w)
	d->tex.data = memdupe(s->tex.data,s->tex.w * s->tex.h * 3);

switch(s->type) {
case st_poly:
	for(i=0; i<s->shp.poly.facec; i++)
		d->shp.poly.pt[i] = memdupe(s->shp.poly.pt[i],
					    sizeof(point)*s->shp.poly.ptc[i]);
	break;

case st_coll:
	d->shp.coll.head = NULL;
	for(sp=s->shp.coll.head; sp; sp=sp->next) {
		shape *cpy = deep_copy_shape(sp);

		cpy->next = d->shp.coll.head;
		d->shp.coll.head = cpy;
		}
	break;

	} /* end switch */
return d;
}


/* unname_shape
 * Clear the name from a shape and all it's children
 */
void unname_shape(shape *s)
{
s->name[0] = 0;
if (s->type == st_copy) {
	shape *sp;

	for(sp=s->shp.coll.head; sp; sp=sp->next)
		unname_shape(sp);
	}
}


/* shape_box
 * Computes the bounding box and complexity of a shape. BUG - doesn't handle
 * rotated shapes properly yet.
 */
int shape_box(shape *sh, double *x1p, double *y1p, double *z1p,
			 double *x2p, double *y2p, double *z2p)
{
int complex = 1;
double x1,y1,z1, x2,y2,z2;

/* Compute basic bounding box
 */
x1 = y1 = z1 = 1000000;
x2 = y2 = z2 = -1000000;
switch(sh->type) {
case st_poly: {
	int i;
	for(i=0; i<sh->shp.poly.facec; i++) {
		int j;
		for(j=0; j<sh->shp.poly.ptc[i]; j++) {
			point *p = &sh->shp.poly.pt[i][j];

			x1 = min(x1,p->x); x2 = max(x2,p->x);
			y1 = min(y1,p->y); y2 = max(y2,p->y);
			z1 = min(z1,p->z); z2 = max(z2,p->z);
			}
		}
	break;
	}
case st_sphere: {
	x1 = sh->shp.sphere.cen.x - sh->shp.sphere.rad;
	y1 = sh->shp.sphere.cen.y - sh->shp.sphere.rad;
	z1 = sh->shp.sphere.cen.z - sh->shp.sphere.rad;
	x2 = sh->shp.sphere.cen.x + sh->shp.sphere.rad;
	y2 = sh->shp.sphere.cen.y + sh->shp.sphere.rad;
	z2 = sh->shp.sphere.cen.z + sh->shp.sphere.rad;
	break;
	}
case st_box: {
	x1 = min(sh->shp.box.c1.x , sh->shp.box.c2.x);
	y1 = min(sh->shp.box.c1.y , sh->shp.box.c2.y);
	z1 = min(sh->shp.box.c1.z , sh->shp.box.c2.z);
	x2 = max(sh->shp.box.c1.x , sh->shp.box.c2.x);
	y2 = max(sh->shp.box.c1.y , sh->shp.box.c2.y);
	z2 = max(sh->shp.box.c1.z , sh->shp.box.c2.z);
	break;
	}
case st_cone: {
	x1 = sh->shp.cone.base.x - sh->shp.cone.rad;
	y1 = sh->shp.cone.base.y - sh->shp.cone.rad;
	z1 = sh->shp.cone.base.z - sh->shp.cone.height/2;
	x2 = sh->shp.cone.base.x + sh->shp.cone.rad;
	y2 = sh->shp.cone.base.y + sh->shp.cone.rad;
	z2 = sh->shp.cone.base.z + sh->shp.cone.height/2;
	break;
	}
case st_tube: {
	x1 = sh->shp.tube.base.x - sh->shp.tube.rad;
	y1 = sh->shp.tube.base.y - sh->shp.tube.rad;
	z1 = sh->shp.tube.base.z - sh->shp.tube.height/2;
	x2 = sh->shp.tube.base.x + sh->shp.tube.rad;
	y2 = sh->shp.tube.base.y + sh->shp.tube.rad;
	z2 = sh->shp.tube.base.z + sh->shp.tube.height/2;
	break;
	}
case st_coll: {
	shape *sp;

	complex = 0;
	for(sp=sh->shp.coll.head; sp; sp=sp->next) {
		double sx1,sy1,sz1, sx2,sy2,sz2;

		complex += shape_box(sp,&sx1,&sy1,&sz1,&sx2,&sy2,&sz2);
		x1 = min(x1,sx1); y1 = min(y1,sy1); z1 = min(z1,sz1);
		x2 = max(x2,sx2); y2 = max(y2,sy2); z2 = max(z2,sz2);
		}
	break;
	}
case st_disc: {
	x1 = sh->shp.disc.cen.x - sh->shp.disc.rad;
	y1 = sh->shp.disc.cen.y;
	z1 = sh->shp.disc.cen.z - sh->shp.disc.rad;
	x2 = sh->shp.disc.cen.x + sh->shp.disc.rad;
	y2 = sh->shp.disc.cen.y;
	z2 = sh->shp.disc.cen.z + sh->shp.disc.rad;
	break;
	}
case st_copy: {
	complex = shape_box(sh->shp.copy.src,&x1,&y1,&z1,&x2,&y2,&z2);
	break;
	}

	} /* end switch */

/* Adjust for position and scaling
 */
x1 *= sh->scx; x2 *= sh->scx;
y1 *= sh->scy; y2 *= sh->scy;
z1 *= sh->scz; z2 *= sh->scz;
x1 += sh->pos.x; x2 += sh->pos.x;
y1 += sh->pos.y; y2 += sh->pos.y;
z1 += sh->pos.z; z2 += sh->pos.z;

/* Store for return
 */
*x1p = x1; *y1p = y1; *z1p = z1;
*x2p = x2; *y2p = y2; *z2p = z2;
return complex;
}

